import os
from pathlib import Path

import pytest

from avulto import DMM, Path as p


def get_fixture_path(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "fixtures", name)


@pytest.fixture
def dmm() -> DMM:
    return DMM.from_file(get_fixture_path("map1.dmm"))


def test_dmm_pathlib():
    assert DMM.from_file(Path(get_fixture_path("map1.dmm")))


def test_dmm_extents(dmm: DMM):
    assert dmm.size == (10, 10, 1)


def test_dmm_get_object(dmm: DMM):
    assert dmm.tiledef(10, 10, 1).find("/obj/foo") == [0]


def test_dmm_get_objvar(dmm: DMM):
    assert dmm.tiledef(7, 7, 1).find("/obj/foo") == [0]
    assert dmm.tiledef(7, 7, 1).find(p("/obj/foo")) == [0]
    assert dmm.tiledef(7, 7, 1).prefab_var(0, "a") == 4


def test_dmm_tile_eq(dmm: DMM):
    assert dmm.tiledef(7, 7, 1) == dmm.tiledef(7, 7, 1)
    assert dmm.tiledef(1, 1, 1) == dmm.tiledef(1, 2, 1)



import random

from avulto import Dmlist, DME

def fill_with_ones(l: Dmlist):
    if not isinstance(l, Dmlist):
        return l

    final_list = {}

    for key in l.keys():
        if l[key]:
            final_list[key] = l[key]
        else:
            final_list[key] = 1

    return final_list


def pickweight(l):
    total = 0
    item = None
    for item, val in l.items():
        total += val

    total = random.uniform(0, total)
    for item, val in l.items():
        total -= val
        if total <= 0:
            return item

    return None


def pick_weight_recursive(list_to_pick):
    result = pickweight(fill_with_ones(list_to_pick))
    while isinstance(result, Dmlist):
        result = pickweight(fill_with_ones(result))
    return result

def prob(p):
    r = random.randint(1, 100)
    if r <= p:
        return 1
    return 0



import os
from typing import Iterator, Any
import pathlib

class Coord3:
    """A three-dimensional integer coordinate. These are currently only returned by certain DMM properties, and are not constructable."""
    x: int
    y: int
    z: int

class Path:
    """A DM typepath."""

    stem: str
    """The final part of the path."""
    parent: "Path"
    """The parent path."""
    is_root: bool
    """Whether or not the path is `/`."""

    abs: str
    """Returns the *absolute* representation of the path, rooted at `/datum`."""
    rel: str
    """Returns the *relative* or *declared* representation of the path."""

    def __init__(self, value):
        """Returns a new path."""

    def child_of(self, other, strict=False) -> bool:
        """Returns whether the path is a child of `other`.
        
        If `strict` is true, the current path will not be considered a child of
        itself.
        """
    def parent_of(self, other, strict=False) -> bool:
        """Returns whether the path is a parent of `other`.

        If `strict` is true, the current path will not be considered a parent of
        itself.
        """

    def __truediv__(self, other: str|"Path") -> "Path":
        """Return the path with the specified suffix."""

class Tile:
    """An individual map tile definition."""

    area_path: Path
    """Returns the path of the tile's area. Returns only the first area if multiple exist."""

    turf_path: Path
    """Returns the path of the tile's turf. Returns only the first area if multiple exist."""

    def add_path(self, index, path: Path | str):
        """Add a prefab with the given `path` at `index`."""
    def convert(self) -> list[dict]:
        """Convert the tile definition to a Python data structure.

        The result of `convert` is completely disassociated from the map tile
        definition, and modifying it will have no effect on the original tile.
        """
    def del_prefab(self, index: int):
        """Deletes the prefab at `index`."""
    def del_prefab_var(self, index: int, name: str):
        """Deletes the variable `name` from the prefab at `index`."""
    def find(self, prefix: Path | str, exact=False) -> list[int]:
        """
        Return the indexes of the prefabs prefixed with the given path `prefix`.
        """
    def prefab_path(self, index: int) -> Path:
        """Return the path of the prefab at `index`."""
    def prefab_var(self, index: int, name: str) -> Any:
        """
        Return the value of the property `name` on the prefab at `index`.

        Raises an error if the property does not exist. For a method that
        returns a default if the property does not exist, see get_prefab_var.
        """
    def get_prefab_var(self, index: int, name: str, default: Any = None):
        """
        Returns the value of the property `name` on the prefab at `index`. If
        the property does not exist, return `default`.
        """

    def prefab_vars(self, index: int) -> list[str]:
        """
        Return the list of variable names on the prefab at `index`.
        """
    def set_prefab_var(self, index: int, name: str, val):
        """Set the value of the variable `name` to `val` at `index`."""
    def set_path(self, index: int, path: Path | str):
        """Set the path of the prefab at `index` to `path`."""

class DMM:
    """A DMM file."""

    size: Coord3
    """The number of tiles in each of the map's three dimensions."""
    filepath: pathlib.Path
    """The original filename of the DMM."""

    @staticmethod
    def from_file(filename: os.PathLike | str) -> "DMM":
        """Creates a DMM from the given `filename`."""
    def coords(self) -> Iterator[tuple[int, int, int]]:
        """Return an iterator over all possible 3D coordinates in the map."""
    def tiles(self) -> Iterator[Tile]:
        """Return an iterator over all unique tiles in the map."""
    def save_to(self, filename: os.PathLike | str):
        """Saves the DMM to the given `filename`."""
    def tiledef(self, x: int, y: int, z: int) -> Tile:
        """Return the tile definition at coords (`x`, `y`, `z`)."""


class ProcDecl:
    """
    A single proc declaration.
    """

    def walk(self, walker: any):
        """Walks the proc AST with *walker*, calling any `visit_*` method names on *walker* if they exist for AST node types."""


class VarDecl:
    """
    A single variable declaration.
    """

    name: str
    """The name of the variable."""
    declared_type: Path | None
    """The declared type of the variable, if specified."""
    const_val: any | None
    """The variable's value, if it can be evaluated as a constant expression."""


class TypeDecl:
    """
    A single type declaration.
    """

    def proc_names(self, declared=False, modified=False, unmodified=False) -> list[str]:
        """Return a list of proc names for the type declaration."""
    def proc_decls(self, name=None) -> list[ProcDecl]:
        """Return proc declarations for the type. If *name* is set, only return proc declarations with this name."""
    def var_names(self, declared=False, modified=False, unmodified=False) -> list[str]:
        """Return a list of variable names for the type declaration."""
    def var_decl(self, name, parents=True) -> VarDecl:
        """Return the proc declaration for variable *name*. If *parents* is True, check up type path if this type does not have this variable set."""

class DME:
    """
    A representation of a single Dreammaker environment.
    """
    filepath: pathlib.Path
    """The original filename of the DMM."""
    types: dict[Path, TypeDecl]
    """A mapping of type paths to their declarations."""

    @staticmethod
    def from_file(filename: os.PathLike | str, parse_procs: bool = False) -> "DME":
        """Creates a DME from the given `filename`.

        If parse_procs is True, the entire AST of the codebase is traversed.
        This is slower than the default but provides more reflection
        information.
        """
    def typesof(self, prefix: Path | str) -> list[str]:
        """Returns a list of type paths with the given `prefix`."""
    def subtypesof(self, prefix: Path | str) -> list[str]:
        """Returns a list of type paths with the given `prefix`, excluding `prefix` itself."""
    def type_decl(self, path: Path | str) -> TypeDecl:
        """Return the type declaration of the given `path`."""

class Dir:
    """An enumeration of directions used in icons."""

    NORTH: "Dir"
    SOUTH: "Dir"
    EAST: "Dir"
    WEST: "Dir"
    NORTHEAST: "Dir"
    NORTHWEST: "Dir"
    SOUTHEAST: "Dir"
    SOUTHWEST: "Dir"

class Rect:
    left: int
    top: int
    width: int
    height: int


class IconState:
    """
    A single icon state in a DMI file.
    """

    name: str
    """The state name."""

    dirs: list[Dir]
    """The directions available in the icon state."""

    frames: int
    """The number of frames in the icon state."""

    movement: bool
    """Returns whether or not the state is a movement state."""

    delays: list[float]
    """Returns an array of frame delays."""

    rewind: bool
    """Returns whether the icon is a rewind icon."""

class DMI:
    """
    A DMI file.
    """
    filepath: pathlib.Path
    """The original filename of the DMM."""

    icon_width: int
    """The width of icons in the file."""

    icon_height: int
    """The height of icons in the file."""

    @staticmethod
    def from_file(filename: os.PathLike | str) -> "DMI":
        """
        Creates a DMI from the given `filename`.
        """
    def state_names(self) -> list[str]:
        """
        Return a list of strings containing all state names in the file.
        """
    def state(self, name: str) -> IconState:
        """
        Return the icon state with the given `name`. If there are duplicates,
        only the first one is returned. Use `states()` to retrieve duplicates.
        """
    def states(self) -> Iterator[IconState]:
        """
        Iterates over all icon states.
        """
    def data_rgba8(self, rect:Rect) -> bytes:
        """Return the byte data of the spritesheet in 8-bit RGBA."""



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'avulto'
copyright = '2025 Warriorstar Orion'
author = 'Warriorstar Orion'
release = 'v0.1.14'
version = release

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinxext.opengraph',
]

autosummary_generate = True
autosummary_imported_members = True

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'Pillow': ('https://pillow.readthedocs.io/en/stable/', None),
}

templates_path = ['_templates']
exclude_patterns = []

html_theme = 'alabaster'
html_sidebars = {
   '**': ['globaltoc.html', 'sourcelink.html', 'searchbox.html'],
}

ogp_site_url = "https://warriorstar-orion.github.io/avulto/"



from avulto import Path as p, paths


def test_paths():
    assert paths.Area == p("/area")
    assert paths.Turf == p("/turf")
    assert paths.Obj == p("/obj")
    assert paths.Mob == p("/mob")
    assert paths.Datum == p("/datum")
    assert paths.Root == p("/")


def test_isparent():
    assert paths.Root.parent_of(p("/datum"))
    assert p("/obj/foo").parent_of(p("/obj/foo/bar"))
    assert p("/obj").parent_of("/obj/foo")


def test_ischild():
    assert paths.Area.child_of(paths.Root)
    assert p("/obj/foo").child_of("/obj")


def test_concat():
    assert paths.Root / "foo" == p("/foo")


def test_suffix():
    assert p("/obj") / "foo" == p("/obj/foo")


def test_parent():
    assert p("/obj").parent == p("/atom/movable")
    assert p("/area/foo").parent == p("/area")


def test_rel_abs():
    assert p("/obj/foo").rel == "/obj/foo"
    assert p("/obj/foo").abs == "/datum/atom/movable/obj/foo"

    assert p("/matrix").abs == "/matrix"

    assert p("/foo").rel == "/foo"
    assert p("/foo").abs == "/datum/foo"

    # Just like BYOND, we care about declaration format
    assert p("/datum/foo").rel == "/datum/foo"
    assert p("/datum/foo").abs == "/datum/foo"


def test_str_compare():
    assert p("/obj/foo") == "/obj/foo"
    assert p("/datum/atom/movable/obj/foo") == "/obj/foo"
    assert "/obj/foo" == p("/obj/foo")



import os

import pytest

from avulto import ast, DME


def get_fixture_path(name):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "fixtures", name)


@pytest.fixture
def dme() -> DME:
    return DME.from_file(get_fixture_path("testenv.dme"), parse_procs=True)


def test_walker_base(dme: DME):
    class VarAndReturnWalker:
        def visit_Return(self, node, source_info):
            pass
            # print(node, source_info)

        def visit_Expr(self, node, source_info):
            pass
            # print(node, source_info)

        def visit_Call(self, node, source_info):
            pass
            # print(node, source_info)

        def visit_Identifier(self, node, source_info):
            pass
            # print(node, source_info)

    varw = VarAndReturnWalker()
    dme.type_decl("/obj/test_object").proc_decls("var_and_return")[0].walk(varw)

def test_visit_call(dme: DME):
    class CallWalker:
        def __init__(self):
            self.calls = list()

        def visit_Call(self, node, source_info):
            self.calls.append(node)

    walker = CallWalker()
    dme.type_decl("/obj/test_object").proc_decls("test_visit_call")[0].walk(walker)
    assert len(walker.calls) == 2
    assert all([isinstance(call, ast.Expression.Call) for call in walker.calls])
